// need WS window!!!!
// clipping

// OzGKS.c : GKS implementation for windows.
//
//
//    a limited single device implementation of the gks
//    graphic standard to support single devices for ozgis
//
//    refer to the gks standard report.
//    mods
//
//    july 92 - original
//    Feb 93 - text sizes
//    May 93 vector array dynamically allocated!
//    June 93 - reading Pixel value
//	  May 97 - 32 bit MFC version
//	  Nov 97 - text params to save calculation, apply aspect to CLIP
//    May 98 - dont zap clip region
//    Sept 99 - LINUX tcl/Tk version

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "TkOzAPI.h"
#include "OzAPI.h"
#include "OzGKS.h"


static void  SetWindVP();


/// The drawing data

///CgksSegment	segment[NUM_SEGS];  the graphics segments
static int	nowSegment = 0;		// the currently selected segment
static int	nowHandle = 0;		// drawing item handle

//  GKS status 

static int polyInteriorStyle=1;	  //fill area interior style  1=solid,2=pattern,3=hatch
static int polyColourIndex=3;       //fill area colour index

static bool clippingFlag=FALSE;       // indicator, 0=none, 1=clip
static float clippingWindow[4];
static int sizeClip = 0;  		// clipping arrays
float* clipX = NULL;
float* clipY = NULL;
float* workX = NULL;
float* workY = NULL;

static bool pickableFlag=FALSE;       // indicator, 0=none, 1=pickable

static int lineType=0;      // line-type
char markerType=1;      // marker type

static float markerSize = 1.0;	// marker size

int	lineWidth = 1;		// line width

static int lineColourIndex=2;       // polyline colour index
static int markerColourIndex=2;       // polymarker colour index
static int textColourIndex=2;       // text colour index

static int currentFont=1;			// selected font number
static int fontPtrecision=1;       // font precision
static float charExpansion=0.5;  // char expansion factor
static float worldCharHt = 1.0;	// char height in world coords
//static int charHeight = 8;       // current font char width requested
//static int charWidth  = 6;       // current font char height
static int textPath=0;       // text path   0=right, 3=down
static int textAlignHoriz=1;      // text alignment horizontal, 1=left, 2=centre
static int textAlignVert=4;       // text alignment vertical, 1=top, 4=base
char fontFace[NUM_FONTS+1][50];	// last font is for markers



//  windows & viewports - for transforming user coords -> device

bool	needTransform = TRUE;		// if need to convert coords at paint
static float wind[] = {0.0,1.0,0.0,0.75};
static float vp[] = {0.0,1.0,0.0,0.75};
static float userWindow[] = {0.0,1.0,0.0,0.75};
static float userViewport[] = {0.0,1.0,0.0,0.75};
static float wsViewportDevCoords[] = {0.0,639.0,0.0,479.0};
static float wsWindowNDC[] = {0.0,1.0,0.0,0.75};
static float xwvp = 1.0;
static float ywvp = 0.75;
static float xndc = 1.0;
static float yndc = 1.0;
static float xoff = 0.0;
static float yoff = 0.0;
static float aspect = 0.75; // aspect  Y/X 
static int   ySwap = 479;
static bool wantAspect = TRUE;

// window status params

int		iRasterCaps;   /* Raster capabilities		      */
int		iPalSize;      /* Size of Physical palette      */
int		iReserved;		// no reserved in palette
static int     colorres;      // number of bits per pixel
int	xScreen;       /* Screen size  */
int	yScreen;       /* Screen size  */
int	nXBorder;      /* Width of window border	      */
int	nYBorder;      /* Width of window border	      */
int	nXTitle;       /* Width of title bar		      */
int	nYTitle;       /* Height of title bar		   */

// OzAPI


int	noXpix=640;        // Width of the application window      
int	noYpix=480;        // Height of the application window    
/////CRect	rectWin;			 current window extent 


// colour LUT

static float red[PALSIZE];
static float green[PALSIZE];
static float blue[PALSIZE];
char hexColour[PALSIZE][7];

// Tcl / Tk

static char* tclCoordsCmd= NULL;
static int sizeCoordsCmd = 0;


/*************************************************************************


  Called from doc at start 


*****************************************************************/


void	GKSinit_ ( )

{
	int	i;
	int r, g, b;

	// start palette

	for(i=0; i < PALSIZE; i++)
	{
		red[i]   = (float)i / (float)PALSIZE;
		green[i] = red[i];
		blue[i]  = 0.0;
		r = 255.0 * red[i] + 0.5;
		g = 255.0 * green[i] + 0.5;
		b = 255.0 * blue[i] + 0.5;
		sprintf(hexColour[i], "%02X%02X%02X", r, g, b);	
	}

	// intialise fonts

	for (i=0; i<NUM_FONTS; i++)
		strcpy(fontFace[i], "-family Helvetica");
	strcpy (fontFace[NUM_FONTS], "-family Symbol");
}





/************************************************************************

	TRANSFORMATIONS

***********************************************************************/



void  Xw2Dev (float x, int* xdev)
//    ==========================

//    convert x in window to device posn

{
float v;


v     = xndc * (xoff + xwvp * (x-wind[0]) );
*xdev =  v + 0.5;

#ifdef DEBUG
fprintf(stderr,  "Xw2Dev %f %d \n",x,*xdev);
#endif

return;
}

void  Xdev2W (float* x, int xdev)
//    ==========================

//    convert x in  device posn to window

{

*x = xdev / xndc ;
*x = *x - xoff;
*x = *x / xwvp + wind[0];
#ifdef DEBUG
fprintf(stderr,  "Xdev2W %f %d \n",*x,xdev);
#endif
return;
}


void  Yw2Dev (float y, int* ydev)
//    ===========================

//    convert y in window to device posn


{
float v;

v     = yndc * (yoff + ywvp * (y-wind[2]) );
*ydev = ySwap - (v + 0.5);  // invert
#ifdef DEBUG
fprintf(stderr,  "Yw2Dev %f %d \n",y,*ydev);
#endif
return;
}


void  Ydev2W (float* y, int ydev)
//    ==========================

//    convert y in  device posn to window

{
*y = ySwap - ydev;
*y =  *y / yndc;
*y = *y - yoff;
*y = *y / ywvp + wind[2];
#ifdef DEBUG
fprintf(stderr,  "Ydev2W %f %d \n",*y,ydev); 
#endif
return;
}


// return scaling from world to pixels for finding box cursor widths etc

void	World2Pixels(float* xScale, float* yScale)
{
		*xScale = xndc * xwvp;
		*yScale = yndc * ywvp;
}


//    set window & viewport parameters

static void  SetWindVP()
//    =================

{
float  fact,off;
int    pix;
int minpxx,maxpxx,minpxy,maxpxy; // clipping pixels extent

noXpix = (wsViewportDevCoords[1] - wsViewportDevCoords[0]) + 1;
noYpix = (wsViewportDevCoords[3] - wsViewportDevCoords[2]) + 1;

aspect = (float)noYpix / (float)noXpix;
ySwap = noYpix - 1;

#ifdef DEBUG
fprintf(stderr,  "!!!!!SetWindVP \n");
fprintf(stderr,  " noXpix %d  noYpix %d aspect %f ySwap %d\n", 
	noXpix, noYpix, aspect, ySwap);
fprintf(stderr,  "userWindow %f %f %f %f\n",
	userWindow[0],userWindow[1],userWindow[2],userWindow[3]);
fprintf(stderr,  "userViewport %f %f %f %f\n",
	userViewport[0],userViewport[1],userViewport[2],userViewport[3]);
fprintf(stderr,  "wsWindowNDC %f %f %f %f\n",
	wsWindowNDC[0],wsWindowNDC[1],wsWindowNDC[2],
	wsWindowNDC[3]);
fprintf(stderr,  "wsViewportDevCoords %f %f %f %f\n",
	wsViewportDevCoords[0],wsViewportDevCoords[1],wsViewportDevCoords[2],
	wsViewportDevCoords[3]);
#endif

//  clipping window

fact = (userWindow[1]-userWindow[0]) / (userViewport[1]-userViewport[0]);
off = fact * userViewport[0];
clippingWindow[0] = userWindow[0] + off;
off = fact * (wsWindowNDC[1]-userViewport[1]);
clippingWindow[1] = userWindow[1] + off;
 
fact = (userWindow[3]-userWindow[2]) / (userViewport[3]-userViewport[2]);
off = fact * (wsWindowNDC[2]-userViewport[2]);
clippingWindow[2] = userWindow[2] + off;
off = fact * (wsWindowNDC[3]-userViewport[3]);
clippingWindow[3] = userWindow[3] + off;

#ifdef DEBUG
fprintf(stderr,  "clippingWindow %f %f %f %f\n",
	clippingWindow[0],clippingWindow[1],clippingWindow[2],
	clippingWindow[3]);
#endif

//  current windows

MoveFloat (userWindow,wind,4);
MoveFloat (userViewport,vp,4);

#ifdef DEBUG
fprintf(stderr,  "wind %f %f %f %f\n",wind[0],wind[1],wind[2],wind[3]);
fprintf(stderr,  "vp %f %f %f %f\n",vp[0],vp[1],vp[2],vp[3]);
#endif

xwvp = (vp[1]-vp[0]) / (wind[1]-wind[0]);
ywvp = (vp[3]-vp[2]) / (wind[3]-wind[2]) / aspect;
xoff = vp[0] - wsViewportDevCoords[0];
yoff = vp[2] - wsViewportDevCoords[2];
xndc = (wsViewportDevCoords[1]-wsViewportDevCoords[0]) / (wsWindowNDC[1]-wsWindowNDC[0]);
yndc = (wsViewportDevCoords[3]-wsViewportDevCoords[2]) / (wsWindowNDC[3]-wsWindowNDC[2]);

#ifdef DEBUG
fprintf(stderr,  "xwvp,ywvp %f %f xoff,yoff %f %f xndc,yndc %f %f\n",
		xwvp,ywvp,xoff,yoff,xndc,yndc);
#endif

//  drawing pixel limits 


if (clippingFlag == 1) // clipping
      {
      pix = noXpix - 1;
      minpxx = pix * vp[0] + 0.5;
      maxpxx = pix * vp[1] + 0.5;
      pix = noYpix - 1;
      minpxy = pix * vp[2] / aspect + 0.5;
      maxpxy = pix * vp[3] / aspect + 0.5;
 
#ifdef DEBUG
 fprintf(stderr,  "CLIP %d %d %d %d\n",minpxx,minpxy,maxpxx,maxpxy);
#endif
     }
else    // all screen, region is null
      {
      //minpxx = 0;
      //minpxy = 0;
      //maxpxx = noXpix - 1; // any bigger fouls printer
      //maxpxy = noYpix - 1;
      }

return;
}






//$$$$$$$$$$$$$$$$$ general $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$




void MoveFloat (float* in, float* out, int num)
//==========================

{
int i;
for (i=0; i<num; i++)
    out[i] = in[i];

return;
}


//%%%%%%%%%%%%%%%% colours %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



void  set_lut (int lut, float cr, float cg, float cb)
//    ==========================
 
//    set LUT array value

// int  lut;  // colour index, 0,1,....
// float cr;    // red value
// float cg;    // green, (0.0-1.0)
// float cb;    // blue

{

	int r, g, b;

	if (lut>PALSIZE)  // too many entries
   		 return;

	red[lut] = cr;
	green[lut] = cg;
	blue[lut] = cb;

	r = 255.0 * cr + 0.5;
	g = 255.0 * cg + 0.5;
	b = 255.0 * cb + 0.5;
	sprintf(hexColour[lut], "%02X%02X%02X", r, g, b);

	if (lut == 0) // background
	{
		sprintf(tclCmd, ".f.c  configure  -background #%s  ", 
		hexColour[lut]);

		Et_EvalF(Et_Interp, tclCmd); 
	}
	

}


//
// $$$$$$$$$$$$$$$$$$$ control functions$$$$$$$$$$$$$$$$$$$
//
 void gopks_ ()
//    =========================
//
//    open gks
{
	return;
}






 void gclks_ ()
//    ==========================
//
//    close gks
//
{
/*******************
	int	seg;
	for (seg=0; seg<NUM_SEGS; seg++)	// clear segments
		segment[seg].Clear();
	//InitLocator(0, NULL, 0, TRUE);	 kill any cursor

	ZapDrawingObjects();

***********************/
	return;
}


 void gopwk_ (int* wkid, int* conid, int* wtype)
//    ==========================
//
//    open workstation
//      int *wkid;   // workstation ident
//      int *conid;   // connection id (lun)
//      int *wtype;   // workstation type,0=output, 2=input-output
{

return;
}




 void gclwk_ (int* wkid)
//    ==========================
//
//    close workstation
//      int *wkid; // workstation id
//
{
/*****************
	int	seg;
	for (seg=0; seg<NUM_SEGS; seg++)
		segment[seg].Clear();
		*****************/

return;
}


 void gacwk_ (int* wkid)
//    ==========================
//
//    activate workstation
//	int *wkid; // workstation id
//
{


//  set device params


wind[0] = 0.0;
wind[1] = 1.0;
wind[2] = 0.0;
wind[3] = 1.0;
wsWindowNDC[0] = 0.0;
wsWindowNDC[1] = 1.0;
wsWindowNDC[2] = 0.0;
wsWindowNDC[3] = 1.0;

return;
}


 void gdawk_ (int* wkid)
//    ==========================
//
//    deactivate workstation
//		int *wkid;

{

      return;
}


 void gclrwk_ (int* wkid, int* cofl)
//    ==========================
//
//    clear workstation
//	int *wkid;   // workstation ident
//	int *cofl;   // control flag, 0=conditional, 1=always

{
	Et_EvalF(Et_Interp, ".f.c  delete all");
	nowHandle = 0;

	return;
}


 void guwk_ (int* wkid,int* regfl)
//    ==========================
//
//    update workstation
//	int *wkid; // workstation ident
//	int *regfl; // control flag, 0=perform, ress, 1=perform

{


return;
}



//
// $$$$$$$$$$$$$$$$$$$output primitives$$$$$$$$$$$$$$$$$$$
//

static void mallocCoordsCmd(int cmdSize, int numCoords)
{

	if (sizeCoordsCmd == 0)
		tclCoordsCmd = malloc(cmdSize);
	else if (sizeCoordsCmd < cmdSize)
		tclCoordsCmd = realloc(tclCoordsCmd, cmdSize);
	if (!tclCoordsCmd)
		OzFatalError("Cant allocate memory");

	if (sizeClip == 0)
	{
		sizeClip = numCoords + 100;
		clipX = malloc(sizeClip*sizeof(float));
		clipY = malloc(sizeClip*sizeof(float));
		workX = malloc(sizeClip*sizeof(float));
		workY = malloc(sizeClip*sizeof(float));
	}
	else if (sizeClip < (numCoords+100))
	{
		sizeClip = numCoords + 100;
		clipX = realloc(clipX, sizeClip*sizeof(float));
		clipY = realloc(clipY, sizeClip*sizeof(float));
		workX = realloc(workX, sizeClip*sizeof(float));
		workY = realloc(workY, sizeClip*sizeof(float));
	}
	if (!clipX || !clipY || !workX || ! workY)
		OzFatalError("Cant allocate memory");}

 int gpl_ (int* n, float* px, float* py)
//    ==========================
//
//    polyline

//int *n;         // number of points
//float px[100],py[100];  // points in world coords

{
	int numClip;
	int iRet;
	int cmdSize = *n * 8 + 100;

	mallocCoordsCmd(cmdSize, *n);

	if (clippingFlag == 1)
	{
		clipline_(n, px, py, clipX, clipY,  
				&sizeClip, &numClip,
				&wind[0],
				&wind[1],
				&wind[2],
				&wind[3]  );
		if (numClip < 2)
			return 0;
		iRet = OzGpl(&numClip, clipX, clipY);
	}
	else
	{
		iRet = OzGpl(n, px, py);
	}
	return (iRet);
}

 
 int OzGpl (int* n, float* px, float* py)
{
	int handle = nowHandle++;
	int num = *n;
	int i;
	int x, y;
	char str[MAX_STRING];

	
	strcpy(tclCoordsCmd, ".f.c create line"); 
	for (i=0; i<num; i++)
	{
		Xw2Dev(px[i], &x);
		Yw2Dev(py[i], &y);
		sprintf (str, " %3d %3d", x, y);
		strcat(tclCoordsCmd, str); 
	}
	sprintf(str, " -fill #%s -width %f -tags {L%d S%d}", 
		hexColour[lineColourIndex], (float)lineWidth,
		handle, nowSegment);
	strcat(tclCoordsCmd, str); 

	Et_EvalF(Et_Interp, tclCoordsCmd); 


	return (handle);
}


 int gpm_ (int* n, float* px, float* py)
//    ==========================
//
//    polymarker 
//int *n;             // number of points (small!!!)
//float px[100],py[100];  // points in world coords
{

	int num = *n;
	int m;
	int x, y;
	char str[MAX_STRING];
	int cmdSize = num * 8 + 200;	
	int handle = nowHandle++;
	int height;
	char text[2];
	int want;
	int count = 0;

	text[0] = markerType;
	text[1] = '\0';

	mallocCoordsCmd(cmdSize, *n);
	
	
	for (m=0; m<num; m++)
	{	
		want = TRUE;
		if (clippingFlag == 1)
		{
			if (*px < wind[0])  want = FALSE;
			if (*px > wind[1])  want = FALSE;
			if (*py < wind[2])  want = FALSE;
			if (*py > wind[3])  want = FALSE;
		}
		if (want)
		{
			count++;
			strcpy(tclCoordsCmd, ".f.c create text"); 

			Xw2Dev(px[m], &x);
			Yw2Dev(py[m], &y);
			sprintf (str, " %d %d", x, y);
			strcat(tclCoordsCmd, str); 

			sprintf(str, " -fill #%s  -tags {M%d S%d}", 
				hexColour[markerColourIndex],
				handle, nowSegment);

			strcat(tclCoordsCmd, str); 

			strcat(tclCoordsCmd, " -anchor c"); 


			strcat(tclCoordsCmd, " -font {"); 
			strcat(tclCoordsCmd, fontFace[NUM_FONTS]); 

			height = xndc * markerSize + 0.5;
			sprintf (str, " -size %d", height);
			strcat(tclCoordsCmd, str);
 			strcat(tclCoordsCmd, "}"); 

			strcat(tclCoordsCmd, " -text "); 
			strcat(tclCoordsCmd, "\""); 
			strcat(tclCoordsCmd, text); 
			strcat(tclCoordsCmd, "\""); 

			Et_EvalF(Et_Interp, tclCoordsCmd); 
		}
	}

	if (count == 0)
		handle = 0;
	return (handle);
}


 int gtx_ (float* px, float* py, char* string, int lString)
//    ======================================================
//
//    text

//float *px,*py;     // text posn in world coords
//char  *string;    // string of characters

{
	char text[MAX_STRING];
	int handle = nowHandle++;
	int x, y;
	char str[MAX_STRING];
	int i;
	char add[3];
	int width, height;
	int start, end;

	if (clippingFlag == 1)
	{
		if (*px < wind[0])  return 0;
		if (*px > wind[1])  return 0;
		if (*py < wind[2])  return 0;
		if (*py > wind[3])  return 0;
	}

	GnuParamAsIs(text, string, lString); // parameter

	start = 0;
	while (text[start] == ' ') // remove leading blanks
		start++;
	end = strlen(text) - 1;
	while(end > 0 && text[end] == ' ') // trailing
		text[end] = '\0';


	strcpy(tclCmd, ".f.c create text"); 
	Xw2Dev(*px, &x);
	Yw2Dev(*py, &y);
	sprintf (str, " %d %d", x, y);
	strcat(tclCmd, str); 

	sprintf(str, " -fill #%s  -tags {T%d S%d}", 
		hexColour[textColourIndex],
		handle, nowSegment);

	strcat(tclCmd, str); 

	if (textAlignVert == 1) // top
	{
		if (textAlignHoriz == 2) // centre
			strcat(tclCmd, " -anchor n"); 
		else if (textAlignHoriz == 3) // right
			strcat(tclCmd, " -anchor ne"); 
		else // left
			strcat(tclCmd, " -anchor nw");
	}
	else // base
	{
		if (textAlignHoriz == 2) // centre
			strcat(tclCmd, " -anchor s"); 
		else if (textAlignHoriz == 3) // right
			strcat(tclCmd, " -anchor se"); 
		else // left
			strcat(tclCmd, " -anchor sw");
 	}


	height = yndc * ywvp * worldCharHt - 0.5; // bit less
	width = charExpansion * height  - 0.5;  // 1 less for space in between
	strcat(tclCmd, " -font {"); 
	strcat(tclCmd, fontFace[currentFont]); 
	sprintf (str, " -size %d", height);
	strcat(tclCmd, str);
 	strcat(tclCmd, "}"); 

	strcat(tclCmd, " -text "); 
	strcat(tclCmd, "\""); 
	if (textPath == 3) // down
	{
		for (i=start; i<strlen(text); i++)
		{
			add[0] = text[i];
			add[1] = '\n';
			add[2] = '\0';
			strcat(tclCmd, add); 
		}
	}
	else
		strcat(tclCmd, &text[start]); 
	strcat(tclCmd, "\""); 

	Et_EvalF(Et_Interp, tclCmd); 

	return (handle);
}

 void gfatrn_ (short* num,float* px, float* py, short* ix, short* iy)
//    ==========================

//    transform array for poly fill

//short *num;             // number of points
//float px[100],py[100];  // points in world coords
//short ix[100],iy[100];  // output device coords
{
int i,n;
	int val;

	n = *num;
	for (i=0; i<n; i++)
    {
	 Xw2Dev (px[i],&val);
		ix[i] = val;
		Yw2Dev (py[i],&val);
		iy[i] = noYpix - val;
    }

	return;
}

 int gfa_ (int* n, float* px, float* py, int* bnd)
//    ==========================

//    fill area 

//int *n;        // number of points
//float px[100],py[100];  // points in world coords
//int *bnd;      // if boundaries reqd, 0=>no
{
	int numClip;
	int iRet;
	int cmdSize = *n * 8 + 100;
	int i;

	mallocCoordsCmd(cmdSize, *n);

	if (clippingFlag == 1)
	{
		//fprintf(stderr, "BEFORE CLIP %d\n", *n);
		//for (i=0; i<*n; i++)
   		//	fprintf(stderr,"%10.2f %10.2f\n", px[i], py[i]);
		clippoly_(n, px, py, 
				workX, workY, clipX, clipY, 
				&sizeClip, &numClip,
				&wind[0],
				&wind[1],
				&wind[2],
				&wind[3]  );
		//fprintf(stderr, "AFTER CLIP %d\n", numClip);
		if (numClip < 3)
			return 0;
		//for (i=0; i<numClip; i++)
   		//	fprintf(stderr,"%10.2f %10.2f\n", clipX[i], clipY[i]);
		iRet = OzGfa(&numClip, clipX, clipY, bnd);
	}
	else
	{
		iRet = OzGfa(n, px, py, bnd);
	}
	return (iRet);
}

 int OzGfa(int* n, float* px, float* py, int* bnd)

{

	bool 	doBoundary = *bnd;
	int handle = nowHandle++;
	int num = *n;
	int i;
	int x, y;
	char str[MAX_STRING];

	
	strcpy(tclCoordsCmd, ".f.c create polygon"); 
	for (i=0; i<num; i++)
	{
		Xw2Dev(px[i], &x);
		Yw2Dev(py[i], &y);
		sprintf (str, " %3d %3d", x, y);
		strcat(tclCoordsCmd, str); 
	}
	sprintf(str, " -fill #%s -width %f -tags {P%d S%d}", 
		hexColour[polyColourIndex], (float)lineWidth,
		handle, nowSegment);
	strcat(tclCoordsCmd, str); 
	if (doBoundary)
	{
		sprintf(str, " -outline #%s ", hexColour[lineColourIndex]);
		strcat(tclCoordsCmd, str); 
	}
	Et_EvalF(Et_Interp, tclCoordsCmd); 
	return (handle);

}

 int  gfrect_ (float* x1, float* x2, float* y1, float* y2)
//    ==============================

//    fill rectangle - handle as POLYGON

//float *x1,*x2,*y1,*y2; // corners in world coords

{
	int	handle;
	float	x[5], y[5];
	int	num;
	int	bnd;

	num = 5;
	bnd = FALSE;
	x[0] = *x1;
	y[0] = *y1;
	x[1] = *x1;
	y[1] = *y2;
	x[2] = *x2;
	y[2] = *y2;
	x[3] = *x2;
	y[3] = *y1;
	x[4] = *x1;
	y[4] = *y1;

	handle = gfa_ (&num, x, y, &bnd);

	return (handle);
}


 void wsetpix_ (short* ix, short* iy, short* index)
//    ==========================

//    set pixel value for images, starting top left

//short *ix;        // integer*2 X
//short *iy;        // integer*2 Y
//short *index;     // LUT index

{

	return;
}



 void wgetpix_ (short* ix, short* iy, short* index)
//    ==========================

//    get pixel value starting at top left

//short *ix;        // integer*2 X
//short *iy;        // integer*2 Y
//short *index;     // LUT index

{

	*index = 0; //?????????
	return;
}

// %%%%%%%%%%%%% update primitives

 void gupdpl_ (int* handle, int* type, int* width, int* colIndex)
//    ==========================
//
//    update polyline


{
	sprintf(tclCmd, ".f.c  itemconfigure  L%d -fill #%s -width %f ", 
		*handle, 
		hexColour[*colIndex], (float)(*width));

	Et_EvalF(Et_Interp, tclCmd); 
}


 void gupdpm_ (int* handle, float* size, int* colIndex)
//    ============================================================
//
//    update polymarker 

{
	int height;

	height = xndc * (*size) + 0.5;

	sprintf(tclCmd, ".f.c  itemconfigure  M%d -fill #%s -font {%s -size %d}", 
		*handle, 
		hexColour[*colIndex], fontFace[NUM_FONTS], height);

	Et_EvalF(Et_Interp, tclCmd); 
}




 void gupdfa_ (int* handle, int* colIndex)
//    ==================================================

//    update fill area 
{
	sprintf(tclCmd, ".f.c  itemconfigure  P%d -fill #%s  ", 
		*handle, 
		hexColour[*colIndex]);

	Et_EvalF(Et_Interp, tclCmd); 
}

//&&&&&&&&&& find primitive attributes

 void gfacol_ (short* index)
//    ====================================

//    return poly fill colour index

//short  *index;

{
	*index = polyColourIndex;
	return;
}


 void gplcol_ (short* index)
//    ====================================

//    return polyline colour index

//short  *index;

{
	*index = lineColourIndex;
	return;
}

/*-----------------------------
 void GFACLP (minx,maxx,miny,maxy)
//    ====================================

//    return clipping rectangle

short  *minx;
short  *maxx;
short  *miny;
short  *maxy;
{
*minx = minpxx;
*miny = minpxy;
*maxx = maxpxx;
*maxy = maxpxy;
return;
}
--------------------------------------*/

// $$$$$$$$$$$$$$$$$$$output primitives$$$$$$$$$$$$$$$$$$$

 void gsseg_ (int* seg)
//    ==========================

//    set current segment

{
	nowSegment = *seg;
	if (nowSegment < 1 || nowSegment > NUM_SEGS)
		OzFatalError ("Illegal segment number");

	return;
}


 void gsln_ (int* ltype)
//    ==========================

//    set line-type range 1 to 5 (1 = solid)
//	  maps to PS_SOLID=0, etc

//int *ltype; // linetype

{

	lineType = *ltype - 1;
	if (lineType < 0 || lineType > 4)
		lineType = 0;
	return;
}


 void gslw_ (int* width)
//    ==========================

//    set polyline width

//int *width;  // width 1,2,3,4
{

	lineWidth = *width;
	return;
}


 void gsplci_ (int* coli)
//    ==========================

//    set polyline colour index

//int *coli;  // colour index

{

	lineColourIndex = *coli;
	return;
}


 void gsmk_ (int* mtype)
//    ==========================

//    set marker type

//int *mtype; // marker type

{

	markerType = *mtype;
	return;
}

 void gsmksize_ (float* msize)
//    ==========================

//    set marker type

//int *mtype; // marker type

{

	markerSize = *msize;
	return;
}


 void gspmci_ (int* coli)
//    ==========================

//    set polymarker colour index

//int *coli; // colour index

{

	markerColourIndex = *coli;
	return;
}

 void winfnt_ (char* font_name, int* id, int lName)
//    ==============================================

//    give font name

//char *font_name;
//int *id;

{
	currentFont = *id;
	if (currentFont < 1 || currentFont > NUM_FONTS)
		OzFatalError("Font ID invalid");
	currentFont--;
	GnuParam (fontFace[currentFont], font_name, lName);
	return;
}


 void mrkfnt_ (char* font_name, int lName)
//    ==========================

//    give font name for markers

//char *font_name;
//int *id;

{

	GnuParam (fontFace[NUM_FONTS], font_name,lName);
	return;
}


 void gstxfp_ (int* font, int* prec)
//    ==========================

//    set font and precision

//int *font; // text font (ident no)
//int *prec; // precision, 0=string, 1=char, 2=stroke

{
	currentFont = *font;
	if (currentFont<1 || currentFont>NUM_FONTS)
		OzFatalError ("Illegal font number");
	currentFont--;
	fontPtrecision = *prec;

	return;
}



 void gschh_ (float* chh)
//    ==========================

//    set character height

//float *chh; // height in world coords

{

	worldCharHt = *chh;
	return;
}


 void gschxp_ (float* chxp)
//    ==========================

//    set character expansion factor

//float *chxp; // expansion factor x to y

{

	charExpansion = *chxp;  

	return;
}


 void gstxci_ (int* coli)
//    ==========================

//    set text colour index

//int *coli; // colour index

{
	textColourIndex = *coli;
	return;
}


 void gstxp_ (int* txp) 
//    ==========================

//    set text path

//int *txp; // path, 0=right, 3=down

{
	textPath = *txp;
	return;
}


 void gstxal_ (int* txalh, int* txalv)
//    ==========================

//    set text alignment

//int *txalh; // horizontal, 1=left, 2=centre, 3=right
//int *txalv; // vertical, 1=top, 4=base

{
	textAlignHoriz = *txalh;
	textAlignVert  = *txalv;
	return;
}


 void gsfais_ (int* ints)
//    ==========================

//    set fill area interior style

//int *ints; // style, 1=solid,2=diagonal hatch at 45, 3=horiz/vert hatch,
//						4=cross hatch at 45, 5=diagonal hatch at 135 , 6=horizontal hatch
//						7=vertical hatch

{
	polyInteriorStyle = *ints;
	return;
}



 void gsfaci_ (int* coli)
//    ==========================

//    set fill area colour index

//int *coli; // colour index

{

	polyColourIndex = *coli;
	return;
}


 void OLDgsasf (int* lasf)
//    =======================

//    set aspect source flags

//int *lasf; // flags, all individual=1 for colourmap


{
	return;
}



// $$$$$$$$$$$$$$$$$$$ workstation attributes $$$$$$$$$$$$$$$$$$$

 void gscr_ (int* wkid, int* ci, float* cr, float* cg, float* cb)
//    ==========================

//    set colour representation

//int *wkid;
//int *ci; // colour index, 0,1,....
//float *cr; // red value
//float *cg; // green, (0.0-1.0)
//float *cb; // blue

{
	int lut;
	lut = *ci;
	set_lut (lut,*cr,*cg,*cb);
//	if (lut == 0)
//	  dc.SetBkColor(GKScolour(lut));

return;
}



 void gslut_ (int* wkid, float* cr, float* cg, float* cb)
//    ==========================

//    set whole look-up table

//int *wkid;
//float *cr; // red array
//float *cg; // green, (0.0-1.0) array
//float *cb; // blue array

{
int i;

for (i = 0; i < PALSIZE; i++)
   set_lut (i,cr[i],cg[i],cb[i]);

//dc.SetBkColor(GKScolour(0));

return;
}



// $$$$$$$$$$$$$$$$$$$ transformation functions $$$$$$$$$$$$$$$$$$$

 void gswn_ (int* tnr, float* xmin, float* xmax, float* ymin, float* ymax)
//    ==========================

//    set window

//int *tnr; // transformation number
//float *xmin;
//float *xmax;
//float *ymin; // window in world coords
//float *ymax;

{
////	CgksWindow*	pWindow;
	userWindow[0] = *xmin;
	userWindow[1] = *xmax;
	userWindow[2] = *ymin; // store values
	userWindow[3] = *ymax;

/////	pWindow = new CgksWindow (*xmin, *xmax, *ymin, *ymax);
////	segment[nowSegment].Add(pWindow);

	SetWindVP();	// keep current for text size calculation

	return;
}


 void gsvp_ (int* tnr, float* xmin, float* xmax, float* ymin, float* ymax)
//    ==========================

//    set viewport

//int  *tnr;  // transformation number
//float *xmin;
//float *xmax;
//float *ymin; // viewport in normalised device coords
//float *ymax;
{
/////	CgksViewport*	pViewport;

	userViewport[0] = *xmin;
	userViewport[1] = *xmax;
	userViewport[2] = *ymin; // store values
	userViewport[3] = *ymax;
///	pViewport = new CgksViewport (*xmin, *xmax, *ymin, *ymax);
////	segment[nowSegment].Add(pViewport);
	SetWindVP();	// keep current for text size calculation

	return;
}


 void OLDgsvpip (int* tnr, int* rtnr, int* relpri)
//    ==========================

//    set viewport input priority

//int *tnr; // transformation number
//int *rtnr; // reference trans. number
//int *relpri; // relative priority; 0=higher, 1=lower

{

      return;
}


 void OLDgselnt (int* tnr)
//    ==========================

//    set transformation no

//int *tnr; // transformation number

{

      return;
}


 void gsclip_ (int* clsw)
//    ==========================

//    set clipping indicator

//int *clsw;   // indicator, 0=none, 1=clip

{

	clippingFlag = *clsw;

//	segment[nowSegment].SetClip(clippingFlag);	// store in segment

	return;
}



 void gclrseg_ ()
//    ==========================

//    clear current segment


{


	///segment[nowSegment].Clear();

	return;
}


 void gspick_ (int* flag)
//    ==========================

//    set pickable indicator for current segment

//int *flag;   // indicator, 0=none, 1=pickable

{

	pickableFlag = *flag;

///	segment[nowSegment].SetPickable(pickableFlag);	// store in segment

	return;
}


void gswkwn_  (wkid,xmin,xmax,ymin,ymax)
//    ==========================

//    set workstation window

long *wkid; // workstation ident
float *xmin;
float *xmax;
float *ymin; // window, normalised device coords
float *ymax;
{
	wsWindowNDC[0] = *xmin;
	wsWindowNDC[1] = *xmax;
	wsWindowNDC[2] = *ymin; // store values
	wsWindowNDC[3] = *ymax;
	SetWindVP();         // window & vp parameters

	return;
}

void gswkvp_  (wkid,xmin,xmax,ymin,ymax)
//    ==========================

//    set workstation viewport

long  *wkid;    // workstation ident
float *xmin;
float *xmax;
float *ymin; // viewport, device coords
float *ymax;

{

	wsViewportDevCoords[0] = *xmin;
	wsViewportDevCoords[1] = *xmax;
	wsViewportDevCoords[2] = *ymin; // store values
	wsViewportDevCoords[3] = *ymax;

	SetWindVP();       // window & vp parameters

	sprintf(tclCmd, ".f.c configure -width %d -height %d", noXpix, noYpix);
	Et_EvalF(Et_Interp, tclCmd); 
	Et_EvalF(Et_Interp, "pack .sf.l -side left -padx 2 -expand yes -fill both"); 
	Et_EvalF(Et_Interp, "pack .sf -side bottom -fill x -pady 2"); return;
}



// $$$$$$$$$$$$$$$$$$$$$$$$$$ inquiry functions $$$$$$$$$$$$$$$$$$$$$$$$$$


 void gdefn_ (int* nnox, int* nnoy, int* nnlut)
//    ===============================================
//
//    return screen params - no DOS equivalent

//int *nnox;   // no pixels across screen 
//int *nnoy;   // no lines down screen 
//int *nnlut;  // length lut
//
{

	*nnox  = noXpix;
	*nnoy  = noYpix;
	*nnlut = PALSIZE;
	return;
}

 void gqpx_ (int* wkid, float* px, float* py, int* errind, int*coli)
//    ==========================

//    inquire pixel

//int *wkid; // workstation ident
//float *px,*py; // point in world coords
//int *errind; // error indicator, 0=ok
//int *coli; // colour index, returned

{
//	CPaintDC dc(gksWND);
	int x,y;

	Xw2Dev (*px, &x);
	Yw2Dev (*py, &y);

///	*coli = dc.GetPixel (x,y);

*coli = 0;  // FIX!!!!!!!!!!!!!!!!!!

	return;
}


// $$$$$$$$$$$$$$$$$$$ input / cursor functions $$$$$$$$$$$$$$$$$$$

 void ginlc_ (int* wkid, int* type, float* initial, int* handle)
//    ==========================

//    initialise locator. Results return in callback routine.

//int  *wkid;      // workstation ident
//int  *type;      // type of locator
//	0 = none, disable
//	1 = cross hair (point)
//	2 = box cursor (select both corners)
//	3 = circle type, select centre & outside point
//	4 = straight line (select both ends)
//	5 = fixed size box
//	6 = digitise line
// float *initial;  // initial data where required
//	fixed size box - size of box in world coords X then Y size

// note that the window / viewport of the current segment is applied, so the
// segment MUST be chosen before the call.

{
	//InitLocator(*type, initial, *handle, TRUE);

	return;
}


// following are cursor requests that wait for results ONCE
// global data is in cursor routines



 void  grqlc_  (int* wkid, float* px, float* py)
//    ==========================

//    request locator

{
	int x1, y1;

	int err;

	oznotbusy_();

	err = Et_EvalF(Et_Interp, " OzLocCursor .f.c  100 200");

	TclCheck(err);

	sscanf(ozResult, "%d %d", &x1, &y1);
	Xdev2W (px, x1);
	Ydev2W (py, y1);
	return;
}


 void  grqbox_ (int* wkid, float* px1, float* py1, float* px2, float* py2)
//    ======================================================

//    request locator - box type, select both corners

{
	int x1, y1, x2, y2;
	int err;


	oznotbusy_();

	err = Et_EvalF(Et_Interp, " OzBoxCursor .f.c  100 200");

	TclCheck(err);

	sscanf(ozResult, "%d %d %d %d", &x1, &y1, &x2, &y2);
	Xdev2W (px1, x1);
	Ydev2W (py1, y1);
	Xdev2W (px2, x2);
	Ydev2W (py2, y2);
	return;
}


 void  grqcirc_ (int* wkid, float* px, float* py, float* radius)
//    ======================================================

//    request locator - circle type, select centre & outside point

{
	int x1, y1, x2, y2;
	float px1, py1, px2, py2;
	int err;


	oznotbusy_();

	err = Et_EvalF(Et_Interp, " OzCircleCursor .f.c  100 200");

	TclCheck(err);


	sscanf(ozResult, "%d %d %d %d", &x1, &y1, &x2, &y2);
	Xdev2W (&px1, x1);
	Ydev2W (&py1, y1);
	Xdev2W (&px2, x2);
	Ydev2W (&py2, y2);
	*px = px1;
	*py = py1;
	*radius = sqrt((px1-px2)*(px1-px2) + (py1-py2)*(py1-py2));
	return;
}


 void  grqlin_ (int* wkid, float* px1, float* py1, float* px2, float* py2)
//    ======================================================

//    request locator - line, select both ends

{
	int x1, y1, x2, y2;
	int err;


	oznotbusy_();

	err = Et_EvalF(Et_Interp, " OzLineCursor .f.c  100 200");

	TclCheck(err);

	sscanf(ozResult, "%d %d %d %d", &x1, &y1, &x2, &y2);
	Xdev2W (px1, x1);
	Ydev2W (py1, y1);
	Xdev2W (px2, x2);
	Ydev2W (py2, y2);
	return;
}


 void  grqfb_ (int* wkid, float* px1, float* py1, float* px2, float* py2)
//    ======================================================
 
//    request locator - fixed size box

{
	int x1, y1, x2, y2;
	int err;


	oznotbusy_();

	err = Et_EvalF(Et_Interp, " OzFixedBoxCursor .f.c  100 200 150 250");

	TclCheck(err);

	sscanf(ozResult, "%d %d %d %d", &x1, &y1, &x2, &y2);
	Xdev2W (px1, x1);
	Ydev2W (py1, y1);
	Xdev2W (px2, x2);
	Ydev2W (py2, y2);
	return;
}


 void  grqdig_ (int* wkid, float* px, float* py, int* sizearray, int* numdig)
//    ======================================================

//    request locator - digitise a line

{
	int i;
	int num;
	int err;
	int x, y;
	char* ptr;


	oznotbusy_();

	err = Et_EvalF(Et_Interp, " OzDigCursor .f.c  100 200");

	TclCheck(err);

	ptr = strtok(ozResult, " ");
	num = atoi(ptr);
	if (num > *sizearray)
		num = *sizearray;
	*numdig = num;

	for (i=0; i<num; i++)
	{
		ptr = strtok(NULL, " ");
		x = atoi(ptr);
		ptr = strtok(NULL, " ");
		y = atoi(ptr);
		Xdev2W (&px[i], x);
		Ydev2W (&py[i], y);	}



	return;
}
